"use strict";

exports.__esModule = true;
exports.default = void 0;

var _es6Promise = require("es6-promise");

var _transaction = _interopRequireDefault(require("./transaction"));

var _utils = require("../common/utils");

var _constants = require("../common/constants");

var _captureNavigation = require("./capture-navigation");

var _env = require("../env");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var TransactionService = function () {
  function TransactionService(logger, config) {
    this._config = config;
    this._logger = logger;
    this.currentTransaction = undefined;
  }

  var _proto = TransactionService.prototype;

  _proto.ensureCurrentTransaction = function ensureCurrentTransaction(options) {
    if (!options) {
      options = this.createOptions();
    }

    var tr = this.getCurrentTransaction();

    if (tr) {
      return tr;
    } else {
      options.canReuse = true;
      options.managed = true;
      return this.createTransaction(undefined, undefined, options);
    }
  };

  _proto.getCurrentTransaction = function getCurrentTransaction() {
    if (this.currentTransaction && !this.currentTransaction.ended) {
      return this.currentTransaction;
    }
  };

  _proto.setCurrentTransaction = function setCurrentTransaction(value) {
    this.currentTransaction = value;
  };

  _proto.createTransaction = function createTransaction(name, type, options) {
    var tr = new _transaction.default(name, type, options);
    this.setCurrentTransaction(tr);

    if (options.checkBrowserResponsiveness) {
      this.startCounter(tr);
    }

    return tr;
  };

  _proto.startCounter = function startCounter(transaction) {
    transaction.browserResponsivenessCounter = 0;

    var interval = this._config.get('browserResponsivenessInterval');

    if (typeof interval === 'undefined') {
      if (_env.__DEV__) {
        this._logger.debug('browserResponsivenessInterval config is undefined!');
      }

      return;
    }

    var id = setInterval(function () {
      if (transaction.ended) {
        window.clearInterval(id);
      } else {
        transaction.browserResponsivenessCounter++;
      }
    }, interval);
  };

  _proto.createOptions = function createOptions(options) {
    var config = this._config.config;
    var presetOptions = {
      transactionSampleRate: config.transactionSampleRate
    };
    var perfOptions = (0, _utils.extend)(presetOptions, options);

    if (perfOptions.managed) {
      perfOptions = (0, _utils.extend)({
        pageLoadTraceId: config.pageLoadTraceId,
        pageLoadSampled: config.pageLoadSampled,
        pageLoadSpanId: config.pageLoadSpanId,
        pageLoadTransactionName: config.pageLoadTransactionName,
        checkBrowserResponsiveness: config.checkBrowserResponsiveness
      }, perfOptions);
    }

    return perfOptions;
  };

  _proto.startManagedTransaction = function startManagedTransaction(name, type, perfOptions) {
    var tr = this.getCurrentTransaction();

    if (!tr) {
      tr = this.createTransaction(name, type, perfOptions);
    } else if (tr.canReuse() && perfOptions.canReuse) {
      if (_env.__DEV__) {
        this._logger.debug("redefining transaction(" + tr.id + ", " + tr.name + ", " + tr.type + ")", 'to', "(" + name + ", " + type + ")", tr);
      }

      tr.redefine(name, undefined, perfOptions);
    } else {
      if (_env.__DEV__) {
        this._logger.debug("ending previous transaction(" + tr.id + ", " + tr.name + ")", tr);
      }

      tr.end();
      tr = this.createTransaction(name, type, perfOptions);
    }

    tr.captureTimings = true;

    if (type === _constants.PAGE_LOAD) {
      tr.options.checkBrowserResponsiveness = false;

      if (perfOptions.pageLoadTraceId) {
        tr.traceId = perfOptions.pageLoadTraceId;
      }

      if (perfOptions.pageLoadSampled) {
        tr.sampled = perfOptions.pageLoadSampled;
      }

      if (tr.name === _constants.NAME_UNKNOWN && perfOptions.pageLoadTransactionName) {
        tr.name = perfOptions.pageLoadTransactionName;
      }
    }

    return tr;
  };

  _proto.startTransaction = function startTransaction(name, type, options) {
    var _this = this;

    var perfOptions = this.createOptions(options);
    var tr;

    if (perfOptions.managed) {
      tr = this.startManagedTransaction(name, type, perfOptions);
    } else {
      tr = new _transaction.default(name, type, perfOptions);
    }

    tr.onEnd = function () {
      return _this.handleTransactionEnd(tr);
    };

    if (_env.__DEV__) {
      this._logger.debug("startTransaction(" + tr.id + ", " + tr.name + ", " + tr.type + ")");
    }

    this._config.events.send(_constants.TRANSACTION_START, [tr]);

    return tr;
  };

  _proto.handleTransactionEnd = function handleTransactionEnd(tr) {
    var _this2 = this;

    return _es6Promise.Promise.resolve().then(function () {
      var name = tr.name,
          type = tr.type;

      if (_this2.shouldIgnoreTransaction(name)) {
        if (_env.__DEV__) {
          _this2._logger.debug("transaction(" + tr.id + ", " + name + ", " + type + ") is ignored");
        }

        return;
      }

      if (type === _constants.PAGE_LOAD) {
        var pageLoadTransactionName = _this2._config.get('pageLoadTransactionName');

        if (name === _constants.NAME_UNKNOWN && pageLoadTransactionName) {
          tr.name = pageLoadTransactionName;
        }
      }

      (0, _captureNavigation.captureNavigation)(tr);

      _this2.adjustTransactionTime(tr);

      var breakdownMetrics = _this2._config.get('breakdownMetrics');

      if (breakdownMetrics) {
        tr.captureBreakdown();
      }

      _this2._config.events.send(_constants.TRANSACTION_END, [tr]);

      if (_env.__DEV__) {
        _this2._logger.debug("end transaction(" + tr.id + ", " + tr.name + ")", tr);
      }
    }, function (err) {
      if (_env.__DEV__) {
        _this2._logger.debug("error ending transaction(" + tr.id + ", " + tr.name + ")", err);
      }
    });
  };

  _proto.adjustTransactionTime = function adjustTransactionTime(transaction) {
    var spans = transaction.spans;
    var earliestSpan = (0, _utils.getEarliestSpan)(spans);

    if (earliestSpan && earliestSpan._start < transaction._start) {
      transaction._start = earliestSpan._start;
    }

    var latestSpan = (0, _utils.getLatestNonXHRSpan)(spans);

    if (latestSpan && latestSpan._end > transaction._end) {
      transaction._end = latestSpan._end;
    }

    var transactionEnd = transaction._end;

    for (var i = 0; i < spans.length; i++) {
      var span = spans[i];

      if (span._end > transactionEnd) {
        span._end = transactionEnd;
        span.type += '.truncated';
      }

      if (span._start > transactionEnd) {
        span._start = transactionEnd;
      }
    }
  };

  _proto.shouldIgnoreTransaction = function shouldIgnoreTransaction(transactionName) {
    var ignoreList = this._config.get('ignoreTransactions');

    if (ignoreList && ignoreList.length) {
      for (var i = 0; i < ignoreList.length; i++) {
        var element = ignoreList[i];

        if (typeof element.test === 'function') {
          if (element.test(transactionName)) {
            return true;
          }
        } else if (element === transactionName) {
          return true;
        }
      }
    }

    return false;
  };

  _proto.startSpan = function startSpan(name, type, options) {
    var tr = this.ensureCurrentTransaction();

    if (tr) {
      var span = tr.startSpan(name, type, options);

      if (_env.__DEV__) {
        this._logger.debug("startSpan(" + name + ", " + type + ")", "on transaction(" + tr.id + ", " + tr.name + ")");
      }

      return span;
    }
  };

  _proto.addTask = function addTask(taskId) {
    var tr = this.ensureCurrentTransaction();

    if (tr) {
      var taskId = tr.addTask(taskId);

      if (_env.__DEV__) {
        this._logger.debug("addTask(" + taskId + ")", "on transaction(" + tr.id + ", " + tr.name + ")");
      }
    }

    return taskId;
  };

  _proto.removeTask = function removeTask(taskId) {
    var tr = this.getCurrentTransaction();

    if (tr) {
      tr.removeTask(taskId);

      if (_env.__DEV__) {
        this._logger.debug("removeTask(" + taskId + ")", "on transaction(" + tr.id + ", " + tr.name + ")");
      }
    }
  };

  return TransactionService;
}();

var _default = TransactionService;
exports.default = _default;